#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
2020 - Rafael Guerreiro Osorio - Ipea.

pynad - method to_csv

Converte os microdados originais para CSV
os arquivos microdados.pnadc.{anual|trimestral}.json
contêm a lista de arquivos originais convertidos
e são usados como base para definir as atualizações
"""


import json
from io import BytesIO, StringIO, TextIOWrapper
from pathlib import Path, PurePosixPath as PPath
from time import time
from zipfile import ZipFile, ZIP_DEFLATED
from .shared import META, MICRO, _VPOS
from .shared import __nomecsv__, __time__, __is_mirror_file__
from .shared import __delete__


def to_csv(mirrorfile, delimiter=','):
    """
    Gera os arquivos de microdados em formato CSV.

    **mirrofile** é o arquivo com a cópia da PNAD

    **delimiter** é opcional, *default* é vírgula
    """
    # aceitar um Path como argumento e verificar requisitos
    if isinstance(mirrorfile, Path):
        mirrorfile = str(mirrorfile)
    if not __is_mirror_file__(mirrorfile):
        return
    if not __check_metadata__(mirrorfile):
        return

    with ZipFile(mirrorfile, 'r', ZIP_DEFLATED) as archive:
        files = archive.namelist()

    if any(MICRO in Path(file).parts[0] for file in files):
        pnads = __select__(mirrorfile)
    else:
        with ZipFile(mirrorfile, 'r', ZIP_DEFLATED) as archive:
            pnads = [[file, json.loads(archive.read(file))] for file in files
                     if all(tag in file for tag in (META,
                                                    'microdados', 'json'))]

    if pnads:
        start0 = time()
        print('Convertendo microdados originais para CSV:')
        archive = ZipFile(mirrorfile, 'a', ZIP_DEFLATED)
        for pnad in pnads:
            for microdado in pnad[1]:
                __convert__(microdado, archive, delimiter)
            metafile = Path(pnad[0]).name
            nome = str(PPath(MICRO, metafile))
            metafile = str(PPath(META, metafile))
            archive.writestr(nome, archive.read(metafile))
        archive.close()
        print(f'Conversão de arquivos em {__time__(start0)}')

    print('Microdados em csv atualizados')


def __convert__(microdado, archive, delimiter=','):
    """
    Faz a conversão do arquivo.

    dicvars = ((*var*, *start*, *stop*), ...), onde:
     - **var** variable code - usado par nomear as variáveis na primeira linha
     - **start** coluna inicial - começando em 0
     - **stop** coluna depois da final

    O conteúdo dos campos é integralmente preservado,
    exceto os campos em branco ou com . que são transformados
    em strings vazias ''
    """
    nome = __nomecsv__(Path(microdado[0]).name, microdado[3])
    if not nome:
        return
    start = time()
    print(f' - Gerando {nome}', end='... ', flush=True)
    original = BytesIO()
    original.write(archive.read(microdado[2]))
    metadata = TextIOWrapper(original, encoding='utf-8')
    metadata.seek(0)
    meta = json.loads(metadata.read())
    dicvars = [(var,
                meta[var][_VPOS][0] - 1,
                meta[var][_VPOS][1]) for var in meta]

    # ler o arquivo zip com os dados e extrair
    with ZipFile(BytesIO(archive.read(microdado[0]))) as src:
        original = BytesIO()
        original.write(src.read(src.namelist()[0]))

    tgt = StringIO()
    tgt.write(delimiter.join([var[0] for var in dicvars]) + '\n')
    fixwidth = TextIOWrapper(original, encoding='utf-8')
    fixwidth.seek(0)
    for reg in fixwidth:
        csvreg = delimiter.join([reg[var[1]:var[2]].strip(' .')
                                 for var in dicvars])
        tgt.write(csvreg + '\n')

    # acrescentar ao arquivo e remover
    tgt.seek(0)
    archive.writestr(str(PPath(MICRO, nome)), tgt.read())
    print(f'OK! {__time__(start)}')


def __select__(mirrorfile):
    """
    Seleciona arquivos para atualização.

    NOTA:-em dezembro de 2020 os dicionários de variáveis
    foram corrigidos para atualizar a descrição de uma categoria
    de uma variável, sem atualizar as bases de microdados

    Atualizações no dicionário fazem gerar novamente os arquivos csv
    das bases que compartilham o dicionário, pois pode ser mudança em
    posição de leitura por conta de erro nas posições, não nos microdados
    """
    archive = ZipFile(mirrorfile, 'a', ZIP_DEFLATED)
    metas = {Path(item).name: json.loads(archive.read(item))
             for item in archive.namelist()
             if all(tag in item for tag in (META, 'microdados', 'json'))}
    dados = {Path(item).name: json.loads(archive.read(item))
             for item in archive.namelist()
             if all(tag in item for tag in (MICRO, 'microdados', 'json'))
             and META not in item}
    update = []
    # arquivos de registro são sempre atualizados
    drop = [str(PPath(MICRO, dado)) for dado in dados]
    for meta in metas:
        new = [item for item in metas[meta]
               if item not in dados[meta]]
        update.append([meta, new])
        drop.extend([str(PPath(MICRO, __nomecsv__(Path(item[0]).name,
                                                  item[3])))
                     for item in dados[meta]
                     if item not in metas[meta]])
    archive.close()
    if all(pnad[1] == [] for pnad in update):
        update = []
    if update:
        print('Atualizando microdados')
        __delete__(mirrorfile, drop)
    return update


def __check_metadata__(mirrorfile):
    # verifica se metadados já foram criados
    with ZipFile(mirrorfile, 'r') as archive:
        items = archive.namelist()
    if not any(META in Path(item).parts[0] for item in items):
        print('Metadados não encontrados')
        return False
    return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
2020 - Rafael Guerreiro Osorio - Ipea.

pynad - module panels

gera no arquivo ZIP de painéis
arquivos de identificação dos grupos domésticos e indivíduos
nos painéis de domicílios

Methods
-------
    ident(args)  para identificar todos os painéis
    load_panel(args)  para carregar um painel
    ident_panel(args)  para identificar o painel carregado
    view(args)  para visualizar um domicílio do painel
    (específico ou aleatório)
"""


import json
from io import BytesIO, StringIO, TextIOWrapper
from random import choice
from time import time
from pathlib import Path, PurePosixPath as PPath
from zipfile import ZipFile, ZIP_DEFLATED
from .shared import REGPES, REGPID, META
from .shared import __time__, __is_mirror_file__, __delete__


class Person():
    """
    representa um registro de pessoa.

    As instâncias desta classe representam registros de pessoas e
    seus atributos correspondem às variáveis empregadas na identificação,
    especificadas em __load_panel_metadata__(), função que retorna
    **attribs** e **categs** usados para criar uma instância;
    **values** é o registro como lista de valores e os
    códigos dos atributos da pessoa
    """

    def __init__(self, attribs, categs, values):
        for idx, attrib in enumerate(attribs):
            self.__dict__[attrib] = values[idx]
        self.sexcat = categs['sexcat'][str(self.sex)]
        self.stdcat = categs['stdcat'][str(self.std)] + 18 * ' '
        self.pid = 0
        self.cla = 0
        self.grp = 0
        self.ind = 0
        self.igi = 0
        self.igd = ''
        self.igm = ''
        self.iga = ''
        self.grpent = 0
        self.indent = 0

    def __repr__(self):
        """
        Retorna uma string que representa a pessoa.

        Contém os atributos usados na identificação,
        bem como os resultantes
        """
        repres = (f'e{self.ent} {self.sexcat[:3]} '
                  + f'{self.dta}-{self.dtm:02.0f}-{self.dtd:02.0f} '
                  + f'{self.ord:02.0f} {self.stdcat[:18]} '
                  + f'cla {self.cla} grpd {self.grp} ind {self.ind:02.0f}')
        if self.igd:
            repres += f' ({self.age} anos - {self.ano - self.age}-99-99)'
        return repres

    def classificada(self, csvdel=','):
        """
        Retorna o registro csv correspondente a pessoa.

        Para escrita no no arquivo de identificação de um painel
        """
        return csvdel.join([str(atrib) for atrib in (self.ano, self.tri,
                                                     self.upa, self.dom,
                                                     self.paibge, self.ent,
                                                     self.ord, self.pid,
                                                     self.cla, self.grp,
                                                     self.grpent,
                                                     self.ind, self.indent,
                                                     self.igd, self.igm,
                                                     self.iga)])

    def ident(self, estrita):
        """Retorna os atributos para a agregação em indivíduos."""
        idnt = f'{self.sexcat[:3]} '
        idnt += f'{self.dta}-{self.dtm:02.0f}-{self.dtd:02.0f}'
        if estrita:
            idnt += f' {self.ord:02.0f} {self.stdcat[:18]}'
        return idnt


def ident(panelfile, pnadc, delimiter=','):
    """
    Identifica os painéis.

    **panelfile** é uma string com o caminho e nome do arquivo de painel
    gerado por **setup()**

    **pnadc** é uma string com o caminho e nome do arquivo com a cópia
    da PNAD Contínua trimestral

    **delimiter** é opcional, *default* é vírgula
    """
    # aceitar Paths como argumentos e verificar
    if isinstance(panelfile, Path):
        panelfile = str(panelfile)
    if isinstance(pnadc, Path):
        pnadc = str(pnadc)
    with ZipFile(panelfile) as archive:
        files = archive.namelist()
        if not any(REGPES in file for file in files):
            print(f'{Path(panelfile).name} não é válido')
            return False
    if not __is_mirror_file__(pnadc):
        return False

    if any(REGPID in file for file in files):
        panelfiles = __id_select__(panelfile)
    else:
        panelfiles = [file for file in files if all(tag in file
                                                    for tag in (REGPES,
                                                                'csv'))]
    if panelfiles:
        with ZipFile(pnadc) as src:
            metadata = json.loads(src.read(str(PPath(META,
                                                     'variaveis.pnadc.'
                                                     + 'trimestral.json'))))
        archive = ZipFile(panelfile, 'a', ZIP_DEFLATED)
        start = time()
        print('Identificando painéis:')
        panelfiles.sort()
        for pfile in panelfiles:
            pid, panel = load_panel(archive, pfile, metadata, delimiter)
            ident_panel(pid, panel, archive)
        archive.writestr(f'{REGPID}/microdados.pnadc.paineis.json',
                         archive.read(f'{REGPES}/'
                                      + 'microdados.pnadc.paineis.json'))
        archive.close()
        print(f'{len(panelfiles)} painéis identificados',
              f'em{__time__(start)}')
    print('Identificação dos painéis atualizada')
    return True


def ident_panel(pid, panel, archive=None, delimiter=','):
    """
    Identifica um painel na memória.

    **pid** e **panel** são retornados por **load_panel()**

    Se especificado o parâmetro **archive**, escreve um
    arquivo csv com a identificação do painel no arquivo
    """
    # prolegômenos
    start = time()
    print(f' - Identificando painel {pid}',
          end='... ', flush=True)

    # passa a lista de pessoas de cada domicílio do painel
    # para o método de identificação __identify__()
    for upa in tuple(panel.keys()):
        for dom in tuple(panel[upa].keys()):
            __identify__(panel[upa][dom])
    print('OK!', __time__(start))

    # escreve o arquivo csv de identificação
    if archive:
        start = time()
        print(f' - Escrevendo arquivo de identificação {pid}',
              end='... ', flush=True)
        target = StringIO()
        target.write(delimiter.join(('ano', 'trimestre', 'upa',
                                     'v1008', 'v1014', 'v1016', 'v2003',
                                     'pid', 'pidcla', 'pidgrp', 'pidgrpent',
                                     'pidind', 'pidindent', 'piddnd',
                                     'piddnm', 'piddna')) + '\n')
        for upa in tuple(panel.keys()):
            for dom in tuple(panel[upa].keys()):
                for pessoa in panel[upa][dom]:
                    target.write(pessoa.classificada() + '\n')
        target.seek(0)
        archive.writestr(f'{REGPID}/pid{pid}.csv', target.read())
        print('OK!', __time__(start))


def load_panel(archive, panelfile, metadata, delimiter=','):
    """
    Carrega um arquivo de microdados de painel.

    **archive** é o arquivo zip com os painéis (gerados por **setup()**)

    **panelfile** é um arquivo de microdados de painel em CSV

    **metadata** é o dicionário de variáveis trimestrasi da PNADC

    **delimiter** é opcional, *default* é vírgula

    Retorna:

    - **pid** código de identificação do painel no formato AAAAT

    - **painel** dicionário python contendo as pessoas dos
      domicílios do painel como instâncias da classe Person, organizados
      pelo número da UPA (chave primária) e pelo número do
      domicílio (v1008, chave secundária)
    """
    def __csv2person__(record):
        pessoa = []
        for attrib in attribs.values():
            pessoa.append(int(record[attrib[1]]))
        return tuple(pessoa)

    # prolegômenos
    start = time()
    print(f' - Carregando {Path(panelfile).name}',
          end='... ', flush=True)

    # carrega o painel
    painel = {}
    pdef = True

    file = TextIOWrapper(BytesIO(archive.read(panelfile)), encoding='utf-8')
    file.seek(0)
    header = file.readline()
    header = header.replace('\n', '').split(delimiter)
    # variáveis usadas na identificação e categorias de sexo e sitdom
    attribs, categs = __load_panel_metadata__(metadata, header)
    for line in file:

        # cada registro de pessoa em texto csv é pŕocessado
        # para gerar uma instância da classe Person, cujos atributos
        # são as variáveis empregadas na identificação
        person = Person(attribs, categs, __csv2person__(line.split(',')))

        # a identificação do painel é feita a partir do
        # primeiro registro de pessoa
        if pdef:
            pid = person.ano * 10 + person.tri
            pdef = False

        # todas as pessoas recebem o código de identificação do painel
        person.pid = pid

        # se for data ignorada, imputa ano de nascimento
        # igual ao ano de inicio do painel
        if person.dtd == 99:
            person.dta = int(person.pid / 10) - person.age

        # a pessoa do registro é inserida na lista de seu domicílio
        # criando as entradas da upa e do domicílio se necessário
        if person.upa not in painel:
            painel[person.upa] = {}
        if person.dom not in painel[person.upa]:
            painel[person.upa][person.dom] = []
        painel[person.upa][person.dom].append(person)
    print('OK!', __time__(start))
    return pid, painel


def view(pid, panel, upa=None, dom=None):
    """
    Retorna uma representação das pessoas de um domicílio.

    **pid** e **panel** são retornados por **load_panel()**

    Se a **upa** e o **dom** não são especificados, escolhe um domicilio
    aleatório do painel
    """
    if not upa:
        upa = choice(tuple(panel.keys()))
    if not dom:
        dom = choice(tuple(panel[upa].keys()))
    print(f'\nPainel {pid}, upa {upa}, domicílio {dom}')
    for pessoa in panel[upa][dom]:
        print(pessoa)


def __load_panel_metadata__(meta, header):
    """Seleciona e identifica os campos do registro csv."""
    # atributos e variáveis correspondentes
    attribs = {'ano': ['ano'],
               'tri': ['trimestre'],
               'upa': ['upa'],
               'dom': ['v1008'],
               'paibge': ['v1014'],
               'ent': ['v1016'],
               'tam': ['v2001'],
               'ord': ['v2003'],
               'std': ['v2005'],
               'sex': ['v2007'],
               'dtd': ['v2008'],
               'dtm': ['v20081'],
               'dta': ['v20082'],
               'age': ['v2009']}

    # acha a posição de cada atributo e obtém as descrições
    # dos códigos de sexo e situação no domicílio
    # tem que ler o header, a ordem das variáveis
    # é diferente
    for attrib in attribs.values():
        attrib.append(header.index(attrib[0]))
    categs = {'stdcat': meta['v2005']['valores'],
              'sexcat': meta['v2007']['valores']}
    return attribs, categs


def __id_select__(panelfile):
    selection = []
    metadata = 'microdados.pnadc.paineis.json'
    with ZipFile(panelfile) as archive:
        items = archive.namelist()
        curpanels = json.loads(archive.read(f'{REGPES}/{metadata}'))
        curpids = json.loads(archive.read(f'{REGPID}/{metadata}'))
    update = [pankey for pankey in curpanels
              if pankey not in curpids
              or curpids[pankey] != curpanels[pankey]]
    if update:
        print('Atualizando a identificação dos painéis')
        selection = [file for file in items for panel in update
                     if panel in file and REGPID in file]
        selection.append(f'{REGPID}/{metadata}')
        __delete__(panelfile, selection)
        selection = [file for file in items
                     for panel in update if panel in file
                     and REGPES in file]
    return selection


def __identify__(pessoas):
    # define a sequencia dos classificadores
    classificadores = (__id_cls1__,
                       __id_cls2__,
                       __id_cls3__)

    # se alguma pessoa tem data de nascimento ignorada, tratamento
    if any(pessoa.dtd == 99 for pessoa in pessoas):
        __id_birthdate_unknown__(pessoas)

    # identificar e separar os grupos domésticos do domicílio
    grpsdom = __id_split_grpds__(pessoas)

    # identificar os indivíduos de cada grupo doméstico
    for grpdom in grpsdom:
        for classificador in classificadores:
            classificadas = classificador(grpsdom[grpdom], grpdom)
            if classificadas:
                break


def __id_aggr_persons__(pessoas, estrita):
    """
    Agrega pessoas em indivíduos.

    **pessoas** são os registros das diferentes pessoas de todas
    as entrevistas realizadas em um domicílio

    **estrita** = True considera sexo, data de nascimento
    número de ordem e situação no domicílio

    **estrita** = False considera sexo e data de nascimento

    Retorna um dicionário de indivíduos com duas listas, a das entrevistas
    do indivíduo, e a das pessoas que lhe correspondem em cada entrevista
    """
    individuos = {}
    for pessoa in pessoas:
        identidade = pessoa.ident(estrita)
        if identidade not in individuos:
            individuos[identidade] = [[], []]
        individuos[identidade][0].append(pessoa.ent)
        individuos[identidade][1].append(pessoa)

    # se apenas sexo e data de nascimento, precisa controlar gemeos,
    # que aparecem como um indivíduo com entrevistas duplicadas
    # é preciso separá-los
    if not estrita:
        for individuo in list(individuos.keys()):
            reflist = list(set(individuos[individuo][0]))
            reflist.sort()
            if individuos[individuo][0] != reflist:
                for entrevista in reflist:
                    gemeos = [idx for idx, item in
                              enumerate(individuos[individuo][0])
                              if item == entrevista]
                    for gemeo, ordem in enumerate(gemeos):
                        identidade = individuo + 'gemeo' + str(gemeo + 1)
                        if identidade not in individuos:
                            individuos[identidade] = [[], []]
                        individuos[identidade][0].append(
                            individuos[individuo][1][ordem].ent)
                        individuos[identidade][1].append(
                            individuos[individuo][1][ordem])
                del individuos[individuo]

    # lista para conjunto de entrevistas do indivíduo
    for individuo in individuos:
        individuos[individuo][0] = list(set(individuos[individuo][0]))
    return individuos


def __id_birthdate_unknown__(pessoas):
    """
    Trata datas de nascimento ignoradas.

    Verifica se uma pessoa com data de nascimento ignorada em uma
    entrevista teve data declarada em outra, imputando a data em caso
    afirmativo
    """
    # separa as pessoas segundo a situação da data de nascimento
    dataign = [pessoa for pessoa in pessoas if pessoa.dtd == 99]
    datacon = [pessoa for pessoa in pessoas if pessoa.dtd != 99]
    for ign in dataign:

        # para cada pessoa com data ignorada, a lista de doadores
        # inicial considera todas as com data conhecida em outras
        # entrevistas, do mesmo sexo e, e com diferença de até
        # 3 anos do ano de nascimento estimado para a data ignorada
        doadores = [pessoa for pessoa in datacon
                    if pessoa.ent != ign.ent
                    and pessoa.sex == ign.sex
                    and abs(pessoa.dta - ign.dta) <= 3]

        # preferência para doadores com a mesma situação no domicílio
        # se existirem, desprezar os demais
        if [pessoa for pessoa in doadores if pessoa.std == ign.std]:
            doadores = [pessoa for pessoa in doadores
                        if pessoa.std == ign.std]

        # se não há doadores com a mesma situação no domicílio,
        # verificar se há doadores do conjunto permitido de troca
        # de código de situação, eliminando os que não estejam
        elif doadores:
            filtro = []
            if ign.std in (1, 2, 3):
                filtro = (1, 2, 3)
            elif ign.std in (4, 5, 6):
                filtro = (4, 5, 6)
            elif ign.std in (8, 9):
                filtro = (8, 9)
            else:
                filtro = (ign.std, )
            doadores = [doador for doador in doadores
                        if doador.std in filtro]

        # havendo doadores, a lista é ordenada pela proximidade
        # do ano de nascimento, sendo escolhido o mais próximo
        # o primeiro da lista
        if doadores:
            difs = [abs(doador.dta - ign.dta) for doador in doadores]
            mindif = [idx for idx, dif in enumerate(difs)
                      if dif == min(difs)][0]
            doador = doadores[mindif]
            ign.igd = doador.dtd
            ign.igm = doador.dtm
            ign.iga = doador.dta

    # imputa as datas de nascimento para as pessoas com data ignorada que
    # encontraram doadoras
    for pessoa in pessoas:
        if pessoa.igd:
            pessoa.dtd = pessoa.igd
            pessoa.dtm = pessoa.igm
            pessoa.dta = pessoa.iga
    return pessoas


def __id_cls1__(pessoas, grpdom):
    """Classificador 1, grupo doméstico sem mudanças."""
    # verifica se o tamanho do grupo doméstico é o mesmo
    # em todas as entrevistas e gera sua lista de entrevistas
    tamanho = pessoas[0].tam
    entdom = []
    for pessoa in pessoas:
        if pessoa.tam != tamanho:
            return []
        if pessoa.ent not in entdom:
            entdom.append(pessoa.ent)
    entdom.sort()

    # agrega as pessoas em indivíduos
    # estrita=True: sexo, data, ordem e sitdom
    individuos = __id_aggr_persons__(pessoas, True)

    # verifica se indivíduos aparecem em todas as entrevistas do grupo
    for individuo in individuos:
        if individuos[individuo][0] != entdom:
            return []

    # identifica as pessoas
    for idx, individuo in enumerate(individuos):
        for pessoa in pessoas:
            identidade = pessoa.ident(True)
            if identidade == individuo:
                pessoa.cla = 1
                pessoa.grp = grpdom
                pessoa.ind = idx + 1
                pessoa.indent = sum(2 ** (item - 1) for item in
                                    individuos[individuo][0])
                pessoa.grpent = sum(2 ** (item - 1) for item in
                                    entdom)
    return pessoas


def __id_cls2__(pessoas, grpdom):
    """Classificador 2 grupos domésticos com tamanho constante e mudanças."""
    # verifica se o tamanho do grupo doméstico é o mesmo
    # em todas as entrevistas e gera sua lista de entrevistas
    tamanho = pessoas[0].tam
    entdom = []
    for pessoa in pessoas:
        if pessoa.tam != tamanho:
            return []
        if pessoa.ent not in entdom:
            entdom.append(pessoa.ent)
    entdom.sort()

    # agrega as pessoas em indivíduos
    # estrita=False: sexo e data
    individuos = __id_aggr_persons__(pessoas, False)

    # verifica se indivíduos aparecem em todas as entrevistas do grupo
    for individuo in individuos:
        if individuos[individuo][0] != entdom:
            return []

    # identifica as pessoas controlando gemeos do mesmo sexo
    indents = {}
    for individuo in individuos:
        indents[individuo] = []
    for idx, individuo in enumerate(individuos):
        for pessoa in pessoas:
            if (pessoa in individuos[individuo][1]
                    and pessoa.ent not in indents[individuo]):
                pessoa.cla = 2
                pessoa.grp = grpdom
                pessoa.ind = idx + 1
                pessoa.indent = sum(2 ** (item - 1) for item in
                                    individuos[individuo][0])
                pessoa.grpent = sum(2 ** (item - 1) for item in
                                    entdom)
                indents[individuo].append(pessoa.ent)
    return pessoas


def __id_cls3__(pessoas, grpdom):
    """Classificador 3 grupos domésticos com mudanças variadas."""
    # recursiva para gerar todas as permutações de individuos
    # que não aparecem na mesma entrevista
    def __permuta__(indids, entrevistas, outras, possiveis):
        for outra in outras:
            entconjunta = entrevistas + individuos[outra][0]
            entconjunta.sort()

            possiveis.append([entconjunta, indids[:] + [outra]])
            if entconjunta != entdom:
                mais = [out for out in outras
                        if not any(ent in entconjunta
                                   for ent in individuos[out][0])]
                if mais:
                    possiveis = __permuta__(indids[:] + [outra],
                                            entconjunta[:], mais,
                                            possiveis)
        return possiveis

    # agrega indívíduos
    def __aggr_ind__(juntar, cla):
        individuos[juntar[1][0]][0] = juntar[0][:]
        for outra in juntar[1][1:]:
            individuos[juntar[1][0]][1].extend(
                individuos[outra][1])
            del individuos[outra]
        for pessoa in individuos[juntar[1][0]][1]:
            pessoa.cla = cla
        return juntar[1]

    # entrevistas do grupo doméstico
    entdom = []
    for pessoa in pessoas:
        if pessoa.ent not in entdom:
            entdom.append(pessoa.ent)
    entdom.sort()

    # agrega as pessoas em indivíduos
    # estrita=False: sexo e data
    individuos = __id_aggr_persons__(pessoas, False)

    # separa e classifica os indivíduos que aparecem em todas as entrevistas
    todas = [individuo for individuo in individuos
             if individuos[individuo][0] == entdom]
    for individuo in todas:
        for pessoa in individuos[individuo][1]:
            pessoa.cla = 3

    # separa e classifica provisoriamente os que aparecem em algumas
    algumas = [individuo for individuo in individuos
               if individuo not in todas]
    for individuo in algumas:
        for pessoa in individuos[individuo][1]:
            pessoa.cla = 4

    # permutações de indivíduos sem aparições conjuntas
    possibilidades = []
    for individuo in algumas:
        entrevistas = list(individuos[individuo][0])
        outras = [outra for outra in list(algumas)
                  if outra != individuo
                  and not any(entrevista in entrevistas
                              for entrevista in individuos[outra][0])]
        if outras:
            possibilidades.extend(__permuta__([individuo],
                                              entrevistas,
                                              outras, []))

    # reduzir permutações a combinações
    organizadas = []
    for possibilidade in possibilidades:
        if not organizadas:
            organizadas.append(possibilidade)
        elif possibilidade not in organizadas:
            adicionar = True
            for organizada in organizadas:
                if (organizada[0] == possibilidade[0]
                        and len(organizada[1]) == len(possibilidade[1])
                        and all(individuo in organizada[1]
                                for individuo in possibilidade[1])):
                    adicionar = False
                    break
            if adicionar:
                organizadas.append(possibilidade)

    # comparar todos os pares de atributos e registrar diferenças
    # a ordem de atribs é relevante
    atribs = ('sex', 'dtd', 'dtm', 'dta', 'std')
    for idx, organizada in enumerate(organizadas):
        organizadas[idx].append([])
        organizadas[idx].append([])
        organizadas[idx].append([])
        possiveis = organizada[1][:]
        while possiveis:
            possivel = possiveis.pop(0)
            pessoa1 = individuos[possivel][1][0]
            outras = [outra for outra in possiveis
                      if outra != possivel]
            for outra in outras:
                pessoa2 = individuos[outra][1][0]
                difs = [atrib for atrib in atribs
                        if pessoa1.__dict__[atrib] !=
                        pessoa2.__dict__[atrib]]
                organizadas[idx][2].append(difs)
                organizadas[idx][3].append([pessoa1.std, pessoa2.std])
                organizadas[idx][4].append([pessoa1.dta, pessoa2.dta])

    # ordenar as combinações para processar primeiro as
    # que resultariam em fusão de mais fragmentos e maior número
    # de entrevistas
    organizadas.sort(key=lambda item: len(item[0]) * 100 + len(item[1]) * 10
                     + 5 - max([len(difs) for difs in item[2]]),
                     reverse=True)

    # conjunto de erros permitidos
    erros = [['sex'],
             ['dtd'],
             ['dtm'],
             ['dta'],
             ['sex', 'dtd'],
             ['sex', 'dtm'],
             ['sex', 'dta'],
             ['dtd', 'dtm'],
             ['dtm', 'dta'],
             ['dtd', 'dta']]

    # com mudança de condição no domicílio
    erstd = [erro[:] + ['std'] for erro in erros]

    # agregar fragmentos de indivíduos
    juntados = []
    residuos = []
    for organizada in organizadas:
        if not any(individuo in organizada[1] for individuo in juntados):

            # fragmentos com mesma condição no domicílio
            # diferença de até 3 anos entre anos de nascimento
            if (all(erro in erros for erro in organizada[2])
                    and all(abs(dta[0] - dta[1]) <= 3
                            for dta in organizada[4])):
                juntados.extend(__aggr_ind__(organizada, 6))

            # fragmentos com diferente condição no domicílio
            # diferença de até 3 anos entre anos de nascimento
            elif (all(erro in erstd for erro in organizada[2])
                  and all(abs(dta[0] - dta[1]) <= 3
                          for dta in organizada[4])):
                juntados.extend(__aggr_ind__(organizada, 7))
            else:
                residuos.append(organizada)
                for individuo in organizada[1]:
                    for pessoa in individuos[individuo][1]:
                        pessoa.cla = 5

    # identifica as pessoas controlando gemeos do mesmo sexo
    indents = {}
    for individuo in individuos:
        indents[individuo] = []
    for idx, individuo in enumerate(individuos):
        for pessoa in pessoas:
            if (pessoa in individuos[individuo][1]
                    and pessoa.ent not in indents[individuo]):
                pessoa.grp = grpdom
                pessoa.ind = idx + 1
                pessoa.indent = sum(2 ** (item - 1) for item in
                                    individuos[individuo][0])
                pessoa.grpent = sum(2 ** (item - 1) for item in
                                    entdom)
                indents[individuo].append(pessoa.ent)
    return pessoas


def __id_split_grpds__(pessoas):
    """Identifica e separa os grupos domésticos de um domicílio."""
    # dicionário de grupos domésticos
    grpsdom = {}

    # separa as pessoas por entrevista
    ents = {}
    for pessoa in pessoas:
        if pessoa.ent not in ents:
            ents[pessoa.ent] = []
        ents[pessoa.ent].append(pessoa)

    # uma entrevista, um grupo; nada a separar...
    if len(ents) == 1:
        grpsdom[1] = pessoas
        for pessoa in grpsdom[1]:
            pessoa.grp = 1
        return grpsdom

    # se mais de uma entrevista
    # agrega as pessoas em indivíduos
    # estrita=False: sexo e, data
    individuos = __id_aggr_persons__(pessoas, False)

    # procurar a entrevista com mais pessoas com data de nascimento conhecida
    inicio = [(ent, len([pessoa for pessoa in ents[ent]
                         if pessoa.dtd < 99]))
              for ent in ents]
    inicio.sort(key=lambda e: e[1], reverse=True)
    ent = inicio[0][0]
    curdom = 0

    # processar entrevistas e ir retirando as atribuídas a um grupo
    # doméstico da lista de entrevistas
    while ents:

        # atribui ao primeiro grupo as pessoas da entrevista inicial
        # e a retira da lista
        curdom += 1
        grpsdom[curdom] = ents[ent]
        del ents[ent]

        # achar outras entrevistas com pessoas de indivíduos que estavam
        # na entrevista inicial e removê-las
        for individuo in individuos.values():
            if any(pessoa in grpsdom[curdom] for pessoa in individuo[1]):
                for ent in individuo[0]:
                    if ent in ents:
                        grpsdom[curdom].extend(ents[ent])
                        del ents[ent]
                    for ini in inicio:
                        if ini[0] == ent:
                            inicio.remove(ini)
                    if not ents:
                        break
            if not ents:
                break

        # verifica se há entrevistas não atribuídas, havendo
        # seleciona a com maior número de pessoas e reinicia o loop
        # criando outro grupo doméstico
        if inicio:
            ent = inicio[0][0]

    # reordenar pessoas
    reord = []
    for grpdom in grpsdom:
        grpsdom[grpdom].sort(key=lambda pessoa: pessoa.ent * 100 + pessoa.ord)
        reord.append([min([pessoa.ent for pessoa in grpsdom[grpdom]]), grpdom])
    reord.sort(key=lambda e: e[0])

    # ordenar grupos domésticos segundo a visita
    # o grupo 1 é o da primeira entrevista
    grpsdomord = {}
    for idx, val in enumerate(reord):
        grpsdomord[idx + 1] = grpsdom[val[1]]
    return grpsdomord



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
2020 - Rafael Guerreiro Osorio - www.ipea.gov.br.

pynad - command line app

script para facilitar o uso da biblioteca pynad
"""

import os
import sys
from pathlib import Path
from time import time
from .shared import PNAD, PNADCA, PNADCT, __time__
from .mirror import Mirror
from .metadata import metadata
from .to_csv import to_csv
from .panels_setup import setup
from .panels_ident import ident
from .panels_reshape import reshape


def main():
    """App linha de comando."""
    if len(sys.argv) == 1 or sys.argv[1].lower() not in COMANDOS:
        print(SYNTAX)
    else:
        if sys.argv[1].lower() == 'pnad':
            COMANDOS[sys.argv[1].lower()](ARQ_PNAD)
        elif sys.argv[1].lower() == 'pnadc':
            COMANDOS[sys.argv[1].lower()](ARQ_PNADC)
        else:
            COMANDOS[sys.argv[1].lower()]()
    print('\n' * 2)


def __identificar__():
    print('\nIdentificar os grupos domésticos e indivíduos dos painéis\n'
          + '---------------------------------------------------------')
    if all(ARQ_PNADC != file.name for file in CURDIR.iterdir()):
        print('Não há cópia da PNAD Contínua no diretório atual:')
        print(f'  {CURDIR}')
        print(SYNTAX)
        return
    ident(Path(CURDIR, ARQ_PAINEIS), Path(CURDIR, ARQ_PNADC))


def __metadados__():
    print('\nOrganizar e extrair metadados\n'
          + '-----------------------------')
    pnads = [file.name for file in CURDIR.iterdir()
             if file.name in (ARQ_PNAD, ARQ_PNADC)]
    if not pnads:
        print('Não há cópias da PNAD ou PNADC no diretório atual:')
        print(f'  {CURDIR}')
        print(SYNTAX)
        return
    if ARQ_PNAD in pnads:
        print('Ainda não implantado para cópias da PNAD')
    if ARQ_PNADC in pnads:
        metadata(Path(CURDIR, ARQ_PNADC))


def __microdados__():
    print('\nGerar microdados em csv\n'
          + '-----------------------')
    pnads = [file.name for file in CURDIR.iterdir()
             if file.name in (ARQ_PNAD, ARQ_PNADC)]
    if not pnads:
        print('Não há cópias da PNAD ou PNADC no diretório atual:')
        print(f'  {CURDIR}')
        print(SYNTAX)
        return
    if ARQ_PNAD in pnads:
        print('Não implantado para cópias da PNAD')
    if ARQ_PNADC in pnads:
        to_csv(Path(CURDIR, ARQ_PNADC))


def __montar__():
    print('\nMontar painéis identificados\n'
          + '----------------------------')
    if all(ARQ_PNADC != file.name for file in CURDIR.iterdir()):
        print('Não há cópia da PNAD Contínua no diretório atual:')
        print(f'  {CURDIR}')
        print(SYNTAX)
        return
    reshape(Path(CURDIR, ARQ_PAINEIS))


def __separar__():
    print('\nSeparar os painéis de domicílios\n'
          + '--------------------------------')
    if all(ARQ_PNADC != file.name for file in CURDIR.iterdir()):
        print('Não há cópia da PNAD Contínua no diretório atual:')
        print(f'  {CURDIR}')
        print(SYNTAX)
        return
    setup(Path(CURDIR, ARQ_PAINEIS), Path(CURDIR, ARQ_PNADC))


def __sincronizar__(arquivo):
    modo = f'Criar {arquivo}'
    if Path(arquivo).is_file():
        modo = f'Sincronizar  {arquivo}'
    print(f'\n{modo}\n{"-" * len(modo)}')
    copia = Mirror(arquivo)
    print('Conectando ao servidor de arquivos do IBGE')

    # ARQ_PNADC deve conter pnadc no nome
    if 'pnadc' in arquivo:
        copia.list_remote_files((PNADCT, PNADCA))
    else:
        copia.list_remote_files(PNAD)
    copia.sync_list()
    dld = [0, 0]
    for pnad in copia.synclst['dld']:
        dld[0] += (len(copia.synclst['dld'][pnad])
                   + len(copia.synclst['rmv'][pnad]))
        dld[1] += sum([file[2] for file in copia.synclst['dld'][pnad]])
    if dld[0]:
        print(f'O arquivo {arquivo} não está sincronizado')
        print(f' - {dld[0]} arquivos estão faltando ou desatualizados')
        print(f' - Tamanho do download: {dld[1] / 1000000:,.1f} MBytes')
        resp = input('Digite sim para baixar os arquivos'
                     + ' (ou Enter para sair): ')
        if resp.lower().strip() not in ('s', 'si', 'sim', 'y', 'yes'):
            if modo == 'criar':
                os.remove(arquivo)
                return
        else:
            copia.sync_now()
    else:
        print(f'O arquivo {arquivo} está sincronizado')


def __tudo__():
    print('\nProcessamento completo da pnadc\n'
          + '-------------------------------')
    print('Conectando ao servidor de arquivos do IBGE')
    start = time()
    copia = Mirror(Path(CURDIR, ARQ_PNADC))
    copia.list_remote_files((PNADCT, PNADCA))
    copia.sync_list()
    copia.sync_now()
    metadata(Path(CURDIR, ARQ_PNADC))
    to_csv(Path(CURDIR, ARQ_PNADC))
    setup(Path(CURDIR, ARQ_PAINEIS), Path(CURDIR, ARQ_PNADC))
    ident(Path(CURDIR, ARQ_PAINEIS), Path(CURDIR, ARQ_PNADC))
    reshape(Path(CURDIR, ARQ_PAINEIS))
    print(f'\nTudo feito em{__time__(start)}')


# -----------------------------------------------
# COMMAND LINE APP USA NOMES PADRÃO PARA ARQUIVOS
# -----------------------------------------------
ARQ_PNAD = 'copia.ibge.pnad.zip'
# deve ter pnadc no nome para __sincronizar__
ARQ_PNADC = 'copia.ibge.pnadc.zip'
ARQ_PAINEIS = 'paineis.ibge.pnadc.zip'
SYNTAX = ('\nUso:  pynad COMANDO'
          + '\n\n  COMANDO pode ser:\n'
          + '\n    pnad        criar/sincronizar uma cópia da PNAD (1976-2015)'
          + '\n    pnadc       criar/sincronizar uma cópia da PNAD Contínua'
          + '\n    metadados   gerar metadados'
          + '\n    microdados  gerar microdados em formato csv'
          + '\n    painel-sep  separar painéis'
          + '\n    painel-id   identificar painéis'
          + '\n    painel-mnt  montar painéis identificados'
          + '\n    pnadc-tudo  pnadc metadata microdata painel-sep|id|mnt'
          + '\n')
COMANDOS = {'pnadc': __sincronizar__,
            'pnad': __sincronizar__,
            'metadados': __metadados__,
            'microdados': __microdados__,
            'painel-sep': __separar__,
            'painel-id': __identificar__,
            'painel-mnt': __montar__,
            'pnadc-tudo': __tudo__}
CURDIR = Path(os.getcwd())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
2020 - Rafael Guerreiro Osorio - Ipea.

pynad - method metadata

Organiza os metadados originais em uma só pasta
e cria versões dos dicionários de variáveis em json
"""


import json
import re
from io import BytesIO
from pathlib import Path, PurePosixPath as PPath
from zipfile import ZipFile, ZIP_DEFLATED
from tablib import Dataset
from .shared import PNAD, PNADCA, PNADCT, ORI, META, TRIDOCS
from .shared import CA_VIS, CA_TRI, CA_DOCS, CA_DATA
from .shared import __is_mirror_file__, __delete__
from .shared import _VPART, _VDESC, _VPER, _VPOS
from .shared import _VSIZE, _VCAT, _VQUES, _MISS


def metadata(mirrorfile):
    """
    Extrai e organiza os metadados.

    Restrito ao necessário para ler e rotular as variáveis e categorias dos
    arquivos de microdados - não inclui
    deflatores, códigos de ocupação, documentos de metodologia etc.
    """
    # aceitar um Path como argumento e verificar
    if isinstance(mirrorfile, Path):
        mirrorfile = str(mirrorfile)
    if not __is_mirror_file__(mirrorfile):
        return
    with ZipFile(mirrorfile, 'a', ZIP_DEFLATED) as archive:
        files = archive.namelist()

    # verificar os dados originais no mirrorfile
    # pode ter qualquer combo de PNAD PNADCT e PNADCA
    metadados = {}
    if any(PNAD == Path(file).parts[1] for file in files):
        metadados[PNAD] = []
    if any(PNADCA == Path(file).parts[1] for file in files):
        metadados[PNADCA] = __metadata_pnadc_anual__(mirrorfile)
    if any(PNADCT == Path(file).parts[1] for file in files):
        metadados[PNADCT] = __metadata_pnadc_trimestral__(mirrorfile)

    if not __atualizados__(mirrorfile, metadados):
        __write_metadata__(mirrorfile, metadados)
        archive.close()
    print('Metadados atualizados')


def __atualizados__(mirrorfile, metadados):
    # carregar microdados json e verificar mudanças em item[0] e [1]
    archive = ZipFile(mirrorfile, 'a', ZIP_DEFLATED)
    files = archive.namelist()
    metanew = {}
    curmetafiles = [Path(item).name for item in files
                    if Path(item).parts[0] == META
                    and 'json' not in item]
    for pnad in metadados:
        metanew[pnad] = [item for item in metadados[pnad]
                         if Path(item[1]).name not in curmetafiles]
        curmicro = pnad.split('_')
        curmicro = f'{META}/microdados.{curmicro[0]}.{curmicro[1]}.json'
        curmicro = json.loads(archive.read(curmicro))
        curmicro = [item[0] for item in curmicro]
        curmicro = [item for item in metadados[pnad]
                    if item[0] not in curmicro]
    archive.close()
    if all(metanew[pnad] == [] for pnad in metadados) and not curmicro:
        return True

    # se houver qualquer atualização, os arquivos
    # microdados*.json terão que ser removidos do mirrorfile
    # isto demora, e aqui, como os metadados atualizados estão
    # todos carregados, apagar todos os arquivos de metadados
    # existentes é mais rápido do que apagar poucos
    curmetafiles = [file for file in files if Path(file).parts[0] == META]
    if curmetafiles:
        print('Atualizando metadados')
        __delete__(mirrorfile, curmetafiles)
    return False


def __write_metadata__(mirrorfile, metadados):
    archive = ZipFile(mirrorfile, 'a', ZIP_DEFLATED)
    for pnad in metadados:
        if pnad == PNAD:
            pass
        if pnad == PNADCA:
            processados = []
            for item in metadados[pnad]:
                target = str(PPath(META, Path(item[1]).name))
                if item[1] not in processados:
                    processados.append(item[1])
                    archive.writestr(target, archive.read(item[1]))
                item[1] = target
                if item[2][0] not in processados:
                    processados.append(item[2][0])
                    archive.writestr(item[2][0], json.dumps(item[2][1]))
                item[2] = item[2][0]
            target = str(PPath(META, 'microdados.pnadc.anual.json'))
            archive.writestr(target, json.dumps(metadados[pnad]))
        if pnad == PNADCT:
            item = metadados[pnad][0]
            item[1] = str(PPath(META, item[1]))
            archive.writestr(item[1], item[2][2].read())
            archive.writestr(item[2][0], json.dumps(item[2][1]))
            for item in metadados[pnad]:
                item[2] = item[2][0]
            target = str(PPath(META, 'microdados.pnadc.trimestral.json'))
            archive.writestr(target, json.dumps(metadados[pnad]))
    archive.close()


def __metadata_pnadc_anual__(mirrorfile):
    """
    Extrai e organiza metadados da pnadc anual.

    Esta rotina depende de como o IBGE organiza os
    arquivos para divulgação

    Em dezembro de 2019 a estrutura de disseminação da pnadc
    anual mudou radicalmente

    Há pastas para:
     - 5 visitas (dicionario por visita e por ano)
     - 4 trimestres (dicionário por trimestre vale para todos os anos)
    """
    archive = ZipFile(mirrorfile, 'a', ZIP_DEFLATED)
    microdados = __metadata_pnadc_anual_arquivos__(archive.namelist())
    for microdado in microdados:
        # lê o dicionário correspondente
        temp = BytesIO(archive.read(microdado[1]))
        doc = Dataset()
        doc.load(temp.getvalue(), format='xls')
        dicionario = __metadata_pnadc_vars__(doc.export('tsv'))

        # acrescenta à lista
        nome = Path(microdado[0]).name
        if 'trimestre' in nome:
            dicid = nome[:-4].split('_')[2]
        else:
            dicid = '.'.join(nome[:-4].split('_')[1:3])
        target = f'variaveis.pnadc.anual.{dicid}.json'
        microdado[2] = [str(PPath(META, target)), dicionario]
    archive.close()
    return microdados


def __metadata_pnadc_anual_arquivos__(zipitems):
    microdados = []
    for vis in range(1, 6):
        datafolder = PPath(ORI, PNADCA, CA_VIS, f'{CA_VIS}_{vis}', CA_DATA)
        docsfolder = PPath(ORI, PNADCA, CA_VIS, f'{CA_VIS}_{vis}', CA_DOCS)
        docfiles = [item for item in zipitems
                    if str(docsfolder) in item
                    and 'dicionario' in item
                    and item.endswith('xls')]
        datafiles = [item for item in zipitems
                     if str(datafolder) in item
                     and item.endswith('zip')]
        for doc in docfiles:
            anodoc = Path(doc).name.split('_')[3]
            microdados.extend([[item, doc, '', PNADCA]
                               for item in datafiles
                               if (Path(item).name.split('_')[1]
                                   == anodoc and f'{CA_VIS.lower()}{vis}'
                                   in item) or
                               (Path(item).name.split('_')[1]
                                in ('2013', '2014') and
                                f'{CA_VIS.lower()}{vis}'
                                in item and anodoc == '2012')])
    for tri in range(1, 5):
        datafolder = PPath(ORI, PNADCA, CA_TRI, f'{CA_TRI}_{tri}', CA_DATA)
        docsfolder = PPath(ORI, PNADCA, CA_TRI, f'{CA_TRI}_{tri}', CA_DOCS)
        docfiles = [item for item in zipitems
                    if str(docsfolder) in item
                    and 'dicionario' in item
                    and item.endswith('xls')]
        datafiles = [item for item in zipitems
                     if str(datafolder) in item
                     and item.endswith('zip')]
        if docfiles:
            doc = docfiles[0]
            microdados.extend([[item, doc, '', PNADCA]
                               for item in datafiles])
    return microdados


def __metadata_pnadc_trimestral__(mirrorfile):
    """
    Extrai e organiza metadados da pnadc trimestral.

    Esta rotina depende de como o IBGE organiza os
    arquivos para divulgação
    """
    archive = ZipFile(mirrorfile, 'a', ZIP_DEFLATED)

    # apenas um dicionário para a pnadc trimestral
    file = [file for file in archive.namelist() if
            all(stub in file for stub
                in (ORI, PNADCT, TRIDOCS, 'Dicionario_e_input'))][0]

    # ler o arquivo zip do dicionário e obter metadados
    buffer = BytesIO(archive.read(file))
    file = ZipFile(buffer)
    diciori = [item for item in file.namelist()
               if 'dicionario' in item][0]
    original = BytesIO(file.read(diciori))

    # le o dicionario com tablib e converte para tsv
    doc = Dataset()
    doc.load(original.getvalue(), format='xls')
    dicionario = __metadata_pnadc_vars__(doc.export('tsv'))
    target = str(PPath(META, 'variaveis.pnadc.trimestral.json'))

    # lista com os arquivos de microdados trimestrais
    microdados = [[item, diciori, [target, dicionario, original], PNADCT]
                  for item in archive.namelist()
                  if PNADCT in item and
                  item.split('/')[2].isnumeric()]
    archive.close()
    return microdados


def __metadata_pnadc_vars__(contents):
    """Depende da estrutura dos dicionários de variáveis."""
    def __clean_label(label):
        label = label.replace('"', '').strip()
        label = label.replace('\n', ' ')
        label = re.sub(' +', ' ', label)
        try:
            label = int(float(label))
        except ValueError:
            pass
        return label

    def __set_bytes(size):
        # dtype bytes para armazenar 9 * colunas da var
        # C signed numeric types
        # 99 - 1 byte
        # 9999 - 2 bytes
        # 999999999 - 4 bytes - float ou int
        # >999999999 - 8 bytes - double ou long
        # pnadc pesos são floats - code 15
        bts = 'ERROR'
        if size <= 2:
            bts = 1
        elif size <= 4:
            bts = 2
        elif size <= 9:
            bts = 4
        elif size <= 14:
            bts = 8
        elif size == 15:
            bts = 15
        return bts

    # meta é o dicionário de variáveis
    meta = {}
    curvar = None

    # seção do questionário
    parte = ''

    # pula linhas de cabeçalho e processa
    rows = contents.split('\r\n')[3:-1]
    for row in rows:

        # line breaks, double spaces e outras coisas
        # limpar campos para processar linhas
        fields = [__clean_label(field) for field in row.split('\t')]

        # linha com informação de "parte" do questionário
        if fields[0] and not fields[1]:
            parte = fields[0].lower()

        # linha principal de variável
        elif all(isinstance(field, int)
                 for field in (fields[0], fields[1])):

            # código (uf, v1008 etc) é a chave em meta
            curvar = fields[2].lower()
            meta[curvar] = {}

            # parte atual
            meta[curvar][_VPART] = parte

            # tuple com:
            # coluna inicial - index em 1
            # coluna final
            # número de colunas
            meta[curvar][_VPOS] = (fields[0],
                                   fields[0] + fields[1] - 1,
                                   fields[1])

            meta[curvar][_VSIZE] = __set_bytes(meta[curvar][_VPOS][2])

            # número do quesito (se tiver)
            meta[curvar][_VQUES] = fields[3]

            # descrição da variável
            meta[curvar][_VDESC] = fields[4].lower()
            if not meta[curvar][_VDESC]:
                meta[curvar][_VDESC] = curvar

            # período
            meta[curvar][_VPER] = fields[7].lower()

            # tem campo 5 - categórica ou info adicional
            meta[curvar][_VCAT] = {}
            especial = (' a ', 'código', 'valor', '130', '01-')
            if (isinstance(fields[5], int)
                    or any(item in fields[5].lower() for item in especial)):
                meta[curvar][_VCAT][fields[5]] = str(fields[6]).lower()
            elif fields[5] or fields[6]:
                meta[curvar][_VCAT] = ', '.join([item.lower()
                                                 for item in fields[5:7]
                                                 if item])

        # linha de categoria
        elif not fields[0] and not fields[1]:
            if not fields[5]:
                fields[5] = _MISS
            try:
                meta[curvar][_VCAT][fields[5]] = fields[6].strip().lower()
            except TypeError:
                print(curvar, fields)
    return meta



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
2020 - Rafael Guerreiro Osorio - Ipea.

pynad - Class Mirror

Gera um arquivo ZIP para armazenar uma cópia local dos
microdados da PNAD ou da PADC e possibilita a sincronização da cópia

A biblioteca zlib deve estar instalada para compressão
"""


import time
from ftplib import FTP, error_perm
from pathlib import Path, PurePosixPath as PPath
from zipfile import ZipFile, ZIP_DEFLATED
from .shared import ORI, URL, FLDRS, ALL
from .shared import __time__, __is_mirror_file__, __delete__


class Mirror():
    """
    Gera ou sincroniza um arquivo zip com cópias locais da PNAD ou PNADC.

    **mirrorfile** é uma string com o caminho e nome do arquivo
    """

    def __init__(self, mirrorfile):
        # aceitar um Path como argumento
        if isinstance(mirrorfile, Path):
            mirrorfile = str(mirrorfile)

        # tenta criar o arquivo ou abri-lo se existente
        try:
            file = ZipFile(mirrorfile, 'x', ZIP_DEFLATED)
        except FileExistsError:
            if __is_mirror_file__(mirrorfile):
                file = ZipFile(mirrorfile, 'a', ZIP_DEFLATED)
            else:
                return

        # se não deu erro de permissão ou outros tipos
        # fecha e guarda o nome
        file.close()
        self.zip = Path(mirrorfile)

        # remotefiles dicionário de listas
        # uma entrada para cada pnad
        # as listas tem o Path relativo ao root
        # atualizado por list_remote_files
        self.remotefiles = {}
        self.distros = {}

        # equivalente com arquivos zip
        self.localfiles = {}
        self.__mapzip__()

        # lista de items para sincronizar
        # atualizado por sync_list
        self.synclst = {'dld': {}, 'rmv': {}}

    @staticmethod
    def __connect__():
        """Conecta ao servidor FTP."""
        server = FTP(URL)
        server.connect()
        server.login()
        return server

    @staticmethod
    def __isfolder__(server, path):
        """Verifica se path é pasta."""
        try:
            server.cwd(path)
            server.cwd('..')
            return True
        except error_perm:
            return False

    def __mapftp__(self, server, pnad, ftproot, ftpcur):
        """Recursiva - lista arquivos in *remotefolder*."""
        server.cwd(str(PPath(ftproot, ftpcur)))
        itens = server.nlst()
        for item in itens:
            sourcename = PPath(ftproot, ftpcur, item)
            if self.__isfolder__(server, str(sourcename)):
                self.__mapftp__(server, pnad, ftproot, PPath(ftpcur, item))
            else:
                self.remotefiles[pnad].append(
                    [ftpcur, item, server.size(str(sourcename))])

    def __mapzip__(self):
        """Lista os arquivos originais no mirrorfile."""
        self.localfiles = {}
        archive = ZipFile(self.zip)
        # aqui é preciso infolist para obter nome e tamanho do arquivo
        files = archive.infolist()
        for file in files:
            path = PPath(file.filename)
            pnad = path.parts[1]
            if pnad not in self.localfiles:
                self.localfiles[pnad] = []
            self.localfiles[pnad].append([PPath(*path.parts[2:-1]),
                                          path.name,
                                          file.file_size])
        archive.close()
        return files

    def list_remote_files(self, pnads=ALL):
        """
        Lista os arquivos de microdados no servidor FTP do IBGE.

        **pnads** é uma string, tuple ou lista com tipos
        definidos pelas constantes:

            - *PNAD*, *PNADCA*, *PNADCT*

        **pnads** determina o tipo de cópia e pode ser qualquer combinação:

            - list_remote_files(*PNAD*) - copia da PNAD
            - list_remote_files((*PNADCA*, *PNADCT*)) - copia
              da PNADC trimestral e anual
            - list_remote_files(ALL) - copia da PNAD e da PNADC
        """
        if not isinstance(pnads, (tuple, list)) and pnads in ALL:
            pnads = (pnads, )
        self.distros = {}
        server = self.__connect__()
        for pnad in pnads:
            print(f' - Verificando arquivos da {pnad} - aguarde',
                  end='... ', flush=True)
            self.remotefiles[pnad] = []
            self.distros[pnad] = {}
            start = time.time()
            self.__mapftp__(server, pnad,
                            PPath('/', *FLDRS[pnad]), PPath(''))
            print(f'OK! {__time__(start)}')
            self.distros[pnad]['files'] = len(self.remotefiles[pnad])
            self.distros[pnad]['size'] = sum([item[2] for item in
                                              self.remotefiles[pnad]])
        server.close()

    def sync_list(self):
        """
        Atualiza *synclst* de acordo com *remote_files*.

        *synclst* contém as listas de arquivos a serem baixados e removidos
        """
        self.__mapzip__()
        for pnad in self.remotefiles:

            # download: arquivos remotos que não existem localmente
            self.synclst['dld'][pnad] = []
            for file in self.remotefiles[pnad]:
                if not self.localfiles:
                    self.synclst['dld'][pnad].append(file)
                elif file not in self.localfiles[pnad]:
                    self.synclst['dld'][pnad].append(file)

            # remove: arquivos locais que não existem remotamente
            self.synclst['rmv'][pnad] = []
            if self.localfiles:
                for file in self.localfiles[pnad]:
                    if file not in self.remotefiles[pnad]:
                        self.synclst['rmv'][pnad].append(file)

    def sync_now(self):
        """Sincroniza de acordo com *synclst*."""

        def __download__(remotefile, localfile):
            with archive.open(localfile, 'w') as handle:
                server.retrbinary(f'RETR {remotefile}', handle.write)

        def __synced__():
            for pnad in self.remotefiles:
                if self.synclst['rmv'][pnad] or self.synclst['dld'][pnad]:
                    return False
            return True

        def __remove__():
            del_lst = []
            for pnad in self.synclst['dld']:
                if self.synclst['dld'][pnad]:
                    dlds = [(file[0], file[1]) for file
                            in self.synclst['dld'][pnad]]
                    for file in self.localfiles[pnad]:
                        if (file[0], file[1]) in dlds:
                            del_lst.append(str(PPath(ORI, pnad,
                                                     file[0], file[1])))
            for pnad in self.synclst['rmv']:
                for file in self.synclst['rmv'][pnad]:
                    del_lst.append(str(PPath(ORI, pnad,
                                             file[0], file[1])))
            return del_lst

        if __synced__():
            print('Cópia local sincronizada')
            return

        start0 = time.time()
        server = self.__connect__()

        # remover arquivos desatualizados
        if self.localfiles:
            remove = __remove__()
            if remove:
                __delete__(self.zip, remove)

        # abre o novo zip - sem os arquivos desatualizados
        archive = ZipFile(self.zip, 'a', ZIP_DEFLATED)

        # baixar
        for pnad in ALL:
            if pnad in self.synclst['dld'] and self.synclst['dld'][pnad]:
                print(f'Arquivos da {pnad}')
                for file in self.synclst['dld'][pnad]:
                    start = time.time()
                    print(f' - Baixando {file[1]}', end='... ', flush=True)
                    source = PPath('/', *FLDRS[pnad], file[0], file[1])
                    target = PPath(ORI, pnad, file[0], file[1])
                    __download__(str(source), str(target))
                    print(f'OK! {__time__(start)}')
        archive.close()
        server.close()
        self.__mapzip__()
        self.sync_list()
        print(f'Arquivos baixados em {__time__(start0)}')
        print('Cópia local sincronizada')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
2020 - Rafael Guerreiro Osorio - www.ipea.gov.br.

pynad - init file
"""
from .mirror import Mirror
from .metadata import metadata
from .shared import PNAD, PNADCT, PNADCA
from .to_csv import to_csv
from .panels_setup import setup
from .panels_ident import ident, load_panel, ident_panel, view
from .panels_reshape import reshape



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
2020 - Rafael Guerreiro Osorio - Ipea.

pynad - method setup

gera um arquivo ZIP com os microdados das pnads
organizados em painéis a partir de cópias locais
da PNAD Contínua Trimestral e Anual
"""


import json
from io import BytesIO, StringIO, TextIOWrapper
from time import time
from pathlib import Path
from zipfile import ZipFile, ZIP_DEFLATED
from .shared import REGPES, PNADCA, PNADCT, MICRO
from .shared import __time__, __is_mirror_file__
from .shared import __delete__, __nomecsv__


def setup(panelfile, pnadc, pnadca='', delimiter=','):
    """
    Separa os microdados por painel.

    **panelfile** é uma string com o caminho e nome do arquivo dos painéis;
    se não existir será criado

    **pnadc** é uma string com o caminho e nome da cópia da PNADC;
    se a pnadc trimestral e a anual estiverem em arquivos separados
    deve ser o da PNADC trimestral

    **pnadca** se a pnadc trimestral e a anual estiverem em arquivos separados
    especifica o caminho e nome do arquivo da PNADC anual

    **delimiter** é opcional, *default* é vírgula
    """
    # aceitar Paths como argumentos e verificar
    if isinstance(panelfile, Path):
        panelfile = str(panelfile)
    if isinstance(pnadc, Path):
        pnadc = str(pnadc)
    if isinstance(pnadca, Path):
        pnadca = str(pnadca)
    if not __is_mirror_file__(pnadc):
        return
    if pnadca and not __is_mirror_file__(pnadca):
        return

    microdados = __set_validate_mirror__(pnadc, pnadca)
    if len(microdados) < 2:
        print('Não foram encontrados os microdados da PNAD Contínua',
              'anual e trimestral com metadados e microdados em formato CSV')
        return False
    if pnadca == '':
        pnadca = pnadc

    start = time()

    # paineis completos disponíveis em mirror - básico e suplementos
    curpanels = __set_panels_mirror__(microdados)
    update = curpanels
    try:
        archive = ZipFile(panelfile, 'x', ZIP_DEFLATED)
        archive.close()
    except FileExistsError:
        with ZipFile(panelfile) as archive:
            files = archive.namelist()
        if not any(REGPES in file for file in files):
            print(f'{Path(panelfile).name} não é válido')
            return False
        update = __set_panels_archive__(panelfile, curpanels)
    if update == 'abort':
        return
    if update:
        __set_join__(update, panelfile, pnadc, pnadca, delimiter)
        with ZipFile(panelfile, 'a', ZIP_DEFLATED) as archive:
            archive.writestr(f'{REGPES}/microdados.pnadc.paineis.json',
                             json.dumps(curpanels))
        print(f'{len(update)} painéis separados',
              f'em{__time__(start)}')
    print('Painéis de domicílios atualizados')


def __set_append__(joinfiles, newheader, delimiter):
    data = {}
    keys = ['upa', 'v1008', 'v2003']
    for visita, bases in enumerate(joinfiles):
        bases[0][1].seek(0)
        header = bases[0][1].readline()
        header = header.replace('\n', '').split(delimiter)
        poskeys = [header.index(var) for var in keys]
        posvars = [header.index(var) if var in header else -1
                   for var in newheader]
        for reg in bases[0][1]:
            reg = reg.replace('\n', '').split(delimiter)
            key = tuple([int(reg[pos]) for pos in poskeys] + [visita])
            if reg[header.index('v1016')] == str(visita + 1):
                data[key] = [reg[pos] if pos > -1 else ''
                             for pos in posvars]
        if len(bases) == 2:
            bases[1][1].seek(0)
            header = bases[1][1].readline()
            header = header.replace('\n', '').split(delimiter)
            poskeys = [header.index(var) for var in keys]
            posvars = [header.index(var) if var in header else -1
                       for var in newheader]
            for reg in bases[1][1]:
                reg = reg.replace('\n', '').split(delimiter)
                key = tuple([int(reg[pos]) for pos in poskeys] + [visita])
                xtr = [reg[pos] if pos > -1 else '' for pos in posvars]
                if key in data:
                    fin = []
                    for idx, _unu in enumerate(xtr):
                        if data[key][idx] == '' and xtr[idx] == '':
                            fin.append('')
                        elif data[key][idx] != '':
                            fin.append(data[key][idx])
                        elif xtr[idx] != '':
                            fin.append(xtr[idx])
                    data[key] = fin
    return data


def __set_join__(panels, panelfile, pnadc, pnadca, delimiter):
    print('Separando painéis:')
    for panel in panels:
        start = time()
        print(f' - Painel {panel}, aguarde', end='... ', flush=True)
        tri = int(str(panel)[-1]) - 1
        ano = panel // 10
        joinfiles = []
        for vis in range(0, 5):
            tri += 1
            if tri > 4:
                tri = 1
                ano = ano + 1
            trimestre = [file for file in panels[panel][1]
                         if f'trimestre{tri}' in file
                         and str(ano) == Path(file).name.split('_')[1]]
            if trimestre:
                joinfiles.append([[__nomecsv__(Path(trimestre[0]).name,
                                               PNADCA), BytesIO()]])
            else:
                joinfiles.append([[__nomecsv__(
                    Path(panels[panel][0][vis]).name,
                    PNADCT), BytesIO()]])
            visita = [file for file in panels[panel][1]
                      if f'visita{vis + 1}' in file]
            if visita:
                joinfiles[-1].append([__nomecsv__(Path(visita[0]).name,
                                                  PNADCA), BytesIO()])
        newheader = __set_join_vars__(joinfiles, pnadc, pnadca, delimiter)
        joined = __set_append__(joinfiles, newheader, delimiter)
        tgt = StringIO()
        tgt.write(delimiter.join(newheader) + '\n')
        for reg in joined.values():
            tgt.write(delimiter.join(reg) + '\n')
        tgt.seek(0)
        with ZipFile(panelfile, 'a', ZIP_DEFLATED) as arc:
            arc.writestr(f'{REGPES}/microdados.pnadc.paineis.{panel}.csv',
                         tgt.read())
        print(f'OK! {__time__(start)}')
        del joinfiles


def __set_join_vars__(joinfiles, pnadc, pnadca, delimiter):
    alphaheader = ['ano', 'trimestre', 'uf', 'capital', 'rm_ride',
                   'estrato', 'posest', 'upa']
    headers = {}
    for visita in joinfiles:
        for file in visita:
            archive = pnadc
            if any(tag in file[0] for tag in ('trimestre', 'visita')):
                archive = pnadca
            with ZipFile(archive) as src:
                file[1].write(src.read(f'{MICRO}/{file[0]}'))
            file[1] = TextIOWrapper(file[1], encoding='utf-8')
            file[1].seek(0)
            headers[file[0]] = file[1].readline().strip('\n').split(delimiter)

    # variáveis começadas por v antes das começadas por s
    ordenar = [[], []]
    for header in headers:
        for var in headers[header]:
            if (var not in alphaheader
                and var not in ordenar[0]
                    and var[0] == 'v'):
                ordenar[0].append(var)
            if (var not in alphaheader
                and var not in ordenar[1]
                    and var[0] == 's'):
                ordenar[1].append(var)
    ordenar[0].sort()
    ordenar[1].sort()
    newheader = alphaheader + ordenar[0] + ordenar[1]
    return newheader


def __set_panels_archive__(panelfile, curpanels):
    with ZipFile(panelfile, 'a', ZIP_DEFLATED) as archive:
        try:
            panels = json.loads(archive.read(
                f'{REGPES}/microdados.pnadc.paineis.json'))
        except KeyError:
            print('Arquivo de painéis inválido - delete e tente de novo')
            return 'abort'
    update = {pankey: panval for pankey, panval in curpanels.items()
              if str(pankey) not in panels
              or panels[str(pankey)] != curpanels[pankey]}
    if update:
        print('Atualização de painéis de domicílios')
        drop = [f'{REGPES}/microdados.pnadc.paineis.{pid}.csv'
                for pid in update if str(pid) in panels]
        drop.append(f'{REGPES}/microdados.pnadc.paineis.json')
        __delete__(panelfile, drop)
    return update


def __set_panels_mirror__(microdados):
    panels = {}
    # para os últimos 4 trimestres não há paineis completos
    # microdados estão ordenados
    for idx, microdado in enumerate(microdados[PNADCT][:-4]):
        ano = int(Path(microdado[0]).name.split('_')[1][2:6])
        tri = int(Path(microdado[0]).name.split('_')[1][:2])
        pid = ano * 10 + tri
        panels[pid] = [[mcdt[0] for mcdt in microdados[PNADCT][idx:idx + 5]]]
        anuais = []
        for seq in range(5):
            if tri > 4:
                tri = 1
                ano += 1
            anuais.append((str(ano), f'trimestre{tri}'))
            anuais.append((str(ano), f'visita{seq + 1}'))
            tri += 1
        panels[pid].append([microdado[0] for microdado in microdados[PNADCA]
                            if any(f'_{anual[0]}_' in microdado[0] and
                                   anual[1] in microdado[0]
                                   for anual in anuais)])
    return panels


def __set_validate_mirror__(pnadc, pnadca):
    micro = {}
    with ZipFile(pnadc) as archive:
        try:
            micro[PNADCT] = json.loads(archive.read(
                f'{MICRO}/microdados.pnadc.trimestral.json'))
        except KeyError:
            pass
    if pnadca:
        with ZipFile(pnadca) as archive:
            try:
                micro[PNADCA] = json.loads(archive.read(
                    f'{MICRO}/microdados.pnadc.anual.json'))
            except KeyError:
                pass
    else:
        with ZipFile(pnadc) as archive:
            try:
                micro[PNADCA] = json.loads(archive.read(
                    f'{MICRO}/microdados.pnadc.anual.json'))
            except KeyError:
                pass
    return micro



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
2020 - Rafael Guerreiro Osorio - Ipea.

pynad - module shared

Constantes e funções privadas usadas por mais de um módulo
"""

import os
from pathlib import Path
from time import time
from subprocess import run
from zipfile import ZipFile, is_zipfile, ZIP_DEFLATED

# não troque as constantes das PNADS...
PNAD = 'pnad'
PNADCA = 'pnadc_anual'
PNADCT = 'pnadc_trimestral'

# microdados no IBGE
URL = 'ftp.ibge.gov.br'
FLDRS = {PNAD: ('Trabalho_e_Rendimento',
                'Pesquisa_Nacional_por_Amostra_de_Domicilios_anual',
                'microdados'),
         PNADCA: ('Trabalho_e_Rendimento',
                  'Pesquisa_Nacional_por_Amostra_de_Domicilios_continua',
                  'Anual', 'Microdados'),
         PNADCT: ('Trabalho_e_Rendimento',
                  'Pesquisa_Nacional_por_Amostra_de_Domicilios_continua',
                  'Trimestral', 'Microdados')}
ALL = (PNAD, PNADCA, PNADCT)

# pastas especificas no IBGE
TRIDOCS = 'Documentacao'
CA_DOCS = 'Documentacao'
CA_DATA = 'Dados'
CA_VIS = 'Visita'
CA_TRI = 'Trimestre'

# nome para pastas das cópias locais e paineis
ORI = 'originais'
META = 'metadados'
MICRO = 'microdados'
REGPES = 'pessoas'
REGIND = 'individuos'
REGPID = 'chaves'

# chaves dos dicionários de variáveis json
_VPART = 'parte'  # parte do registro: identificação e controle etc.
_VDESC = 'desc'  # description
_VPER = 'periodo'  # period
_VPOS = 'colunas'  # position in fixed width microdata file
_VSIZE = 'bytes'  # bytes needed to store the variable
_VCAT = 'valores'  # content - categories
_VQUES = 'quesito'  # question number if from the questionnaire
_MISS = 'vazio'  # key for "não aplicável" in _VCAT


# ------------------------------------
# funções usadas por mais de um módulo
# ------------------------------------


def __time__(start):
    # string com o tempo formatado desde start
    end = time()
    hours = int((end - start) / 3600)
    mins = int(((end - start) - hours * 3600) / 60)
    segs = int((end - start) - hours * 3600 - mins * 60)
    mils = int(((end - start) - int(end - start)) * 1000)
    if hours:
        result = f'{hours:3.0f}h{mins:3.0f}m{segs:3.0f}s'
    elif mins:
        result = f'{mins:3.0f}m{segs:3.0f}s'
    elif segs:
        result = f'{segs:3.0f}s'
    else:
        result = f'{mils:3.0f}ms'
    return result


def __nomecsv__(nome, pnad):
    # nome do arquivo csv a partir do original e tipo de pnad
    if pnad == PNAD:
        pass
    if pnad == PNADCA:
        stub = '.'.join(nome[:-4].split('_')[1:3])
    if pnad == PNADCT:
        stub = nome[:-4].split('_')
        stub = f'{stub[1][2:]}.{stub[1][1]}'
    nome = ".".join(pnad.split('_'))
    nome = f'microdados.{nome}.{stub}.csv'
    return nome


def __is_mirror_file__(mirrorfile):
    # verifica se arquivo zip é um mirror válido
    valid = True
    if not Path(mirrorfile).is_file():
        valid = False
    elif not is_zipfile(mirrorfile):
        valid = False
    if valid:
        with ZipFile(mirrorfile) as src:
            items = src.namelist()
        if not [item for item in items
                if ORI in item and any(pnad in item for pnad in ALL)]:
            valid = False
    if not valid:
        print(f'{mirrorfile} inválido')
    return valid


def __delete__(archive, delete):
    # zipfile é horrível para deletar, usar zip ou 7zip se disponíveis
    def __withzipfile__():
        os.rename(archive, archive + '.old')
        with ZipFile(archive + '.old') as src:
            with ZipFile(archive, 'w', ZIP_DEFLATED) as tgt:
                for file in src.namelist():
                    if file not in delete:
                        tgt.writestr(file, src.read(file))
        os.remove(archive + '.old')
    start = time()
    print(' - Removendo arquivos desatualizados', end='...', flush=True)
    try:
        run(('zip', '-d', archive, *delete),
            capture_output=True, check=True)
    except FileNotFoundError:
        if os.name == 'posix':
            u7z = '7za'
        else:
            u7z = '7z'
        try:
            run((u7z, 'd', archive, *delete),
                capture_output=True, check=True)
        except FileNotFoundError:
            print(' (zipfile) ', end='', flush=True)
            __withzipfile__()
    print(f'OK! {__time__(start)}')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
2020 - Rafael Guerreiro Osorio - Ipea.

pynad
"""

from setuptools import setup, find_packages

with open('README.rst', 'r') as file:
    long_description = file.read()

setup(
    name='pynad',
    version='1.0.2',
    author='Rafael Guerreiro Osorio',
    author_email='rafael.osorio@ipea.gov.br',
    description="An application to manage pnad's microdata and metadata",
    long_description=long_description,

    # homepage
    url='',

    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: '
        + 'GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
    packages=find_packages(),
    install_requires=['tablib', 'xlrd', 'xlwt'],

    # command line app
    entry_points={
        "console_scripts": ['pynad = pynad.pynad:main']
        })



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
2020 - Rafael Guerreiro Osorio - Ipea.

pynad - method reshape

gera um um conjunto de microdados de painéis com os registros
de indivíduos

nesta base, as variáveis dos indivíduos são replicadas por visita
e sufixadas com o número da visita eg. v2005_1, v2005_2 ... v2005_5
"""


import json
from io import BytesIO, StringIO, TextIOWrapper
from pathlib import Path
from time import time
from zipfile import ZipFile, ZIP_DEFLATED
from tablib import Dataset
from .shared import REGIND, REGPID, REGPES
from .shared import __time__, __delete__


def reshape(panelfile, delimiter=','):
    """
    Monta ou atualiza os painéis de individuos identificados.

    **panelfile** arquivo dos painéis

    **delimiter** é opcional, *default* é vírgula
    """
    # aceitar Paths como argumentos e verificar
    if isinstance(panelfile, Path):
        panelfile = str(panelfile)
    with ZipFile(panelfile) as src:
        files = src.namelist()
    if not any(file for file in files if REGPES in file or REGPID in file):
        print(f'{panelfile} inválido')
        return
    panels = __select__(panelfile)
    if panels == 'abort':
        return
    start = time()
    if panels:
        print('Montando painéis identificados')
        for panel in panels:
            print(f'Painel {panel}')
            __reshape_panel__(panel, panelfile, delimiter)
        with ZipFile(panelfile, 'a', ZIP_DEFLATED) as tgt:
            tgt.writestr(f'{REGIND}/microdados.pnadc.paineis.json',
                         tgt.read(f'{REGPID}/microdados.pnadc.paineis.json'))
        print(f'{len(panels)} painéis identificados montados',
              f'em{__time__(start)}')
    print('Painéis de indivíduos atualizados')


def __drop_columns__(datafile):
    data = Dataset()
    datafile.seek(0)
    data.load(datafile, format='csv')
    for var in data.headers[:]:
        if not any(obs != '' for obs in data[var]):
            del data[var]
    return StringIO(data.export('csv'))


def __reshape_panel__(panel, panelfile, delimiter):
    with ZipFile(panelfile) as src:
        stacked = BytesIO(src.read(
                          f'{REGPES}/microdados.pnadc.paineis.{panel}.csv'))
        pidfile = BytesIO(src.read(f'{REGPID}/pid{panel}.csv'))

    header, individuos = __unstack__(stacked, pidfile, delimiter)
    __write_base__(panel, panelfile, header, individuos, delimiter)
    __write_sections__(panel, panelfile, header, individuos, delimiter)


def __select__(panelfile):
    selection = []
    metadata = 'microdados.pnadc.paineis.json'
    with ZipFile(panelfile) as archive:
        files = archive.namelist()
        curpanels = json.loads(archive.read(f'{REGPES}/{metadata}'))
        curpids = json.loads(archive.read(f'{REGPID}/{metadata}'))
        if any(REGIND in file for file in files):
            unstacked = json.loads(archive.read(f'{REGIND}/{metadata}'))
        else:
            unstacked = []
    update = [pankey for pankey in curpanels
              if pankey not in curpids
              or curpids[pankey] != curpanels[pankey]]
    if update:
        print('Há painéis não identificados')
        return 'abort'

    update = [pankey for pankey in curpids
              if pankey not in unstacked
              or curpids[pankey] != unstacked[pankey]]
    if update and unstacked:
        print('Atualização de painéis de indivíduos')
        selection = [file for file in files for panel in update
                     if panel in file and REGIND in file]
        selection.append(f'{REGIND}/{metadata}')
        __delete__(panelfile, selection)
    return update


def __unstack__(stacked, pidfile, delimiter):
    start = time()
    print(' - Carregando pessoas por indivíduo (reshape wide, unstack)',
          end='... ', flush=True)
    individuos = {}
    src = TextIOWrapper(pidfile, encoding='utf-8')
    pidheader = src.readline()
    pidheader = pidheader.replace('\n', '').split(delimiter)
    indpos = tuple(pidheader.index(var) for var in
                   ('upa', 'v1008', 'pidgrp', 'pidind'))
    pespos = tuple(pidheader.index(var) for var in
                   ('upa', 'v1008', 'v1016', 'v2003'))
    pidpos = tuple(pidheader.index(var) for var in pidheader
                   if var[:3] == 'pid')
    for reg in src:
        reg = reg.replace('\n', '').split(delimiter)
        indkey = tuple(int(reg[pos]) for pos in indpos)
        try:
            peskey = tuple(int(reg[pos]) for pos in pespos)
        except IndexError:
            print(pespos, reg)
            raise
        pid_dt = tuple(reg[pos] for pos in pidpos)
        if indkey not in individuos:
            individuos[indkey] = [pid_dt, [peskey]]
        else:
            individuos[indkey][1].append(peskey)

    pidheader = [var for var in pidheader if var[:3] == 'pid']
    pessoas = {}
    src = TextIOWrapper(stacked, encoding='utf-8')
    panheader = src.readline()
    panheader = panheader.replace('\n', '').split(delimiter)
    pespos = tuple(panheader.index(var) for var in
                   ('upa', 'v1008', 'v1016', 'v2003'))
    for reg in src:
        reg = reg.replace('\n', '').split(delimiter)
        peskey = tuple(int(reg[pos]) for pos in pespos)
        pessoas[peskey] = tuple(reg)

    for ind in individuos:
        pesind = []
        for peskey in individuos[ind][1][:]:
            pesind.append(individuos[ind][0] + pessoas[peskey])
        individuos[ind] = pesind
    print('OK!', __time__(start))
    pesheader = __weights__((pidheader + panheader), individuos)
    return pesheader, individuos


def __weights__(header, individuos):
    posv1016 = header.index('v1016')
    posv1027 = header.index('v1027')
    posv1029 = header.index('v1029')
    posposest = header.index('posest')
    popposest = {}
    for vis in range(1, 6):
        popposest[str(vis)] = {}
    for individuo in individuos:
        for pessoa in individuos[individuo]:
            visita = pessoa[posv1016]
            posest = pessoa[posposest]
            if posest not in popposest[visita]:
                popposest[visita][posest] = [int(pessoa[posv1029]),
                                             float(pessoa[posv1027])]
            else:
                popposest[visita][posest][1] += float(pessoa[posv1027])
    for individuo in individuos:
        for idx, pessoa in enumerate(individuos[individuo]):
            visita = pessoa[posv1016]
            posest = pessoa[posposest]
            peso = (popposest[visita][posest][0] /
                    popposest[visita][posest][1] *
                    float(pessoa[posv1027]))
            individuos[individuo][idx] += (str(peso), )
    header.append('pidpeso')
    return header


# constantes para os nomes de blocos - sections
BASE = 'basico'
EDUC = 'educa'
TRAB = 'trabalho'
OREN = 'rendas'
MORA = 'moradia'
TICS = 'tics'
TURI = 'turismo'
DERI = 'derivadas'
TRIN = 'trabinfa'


def __write_base__(panel, panelfile, pesheader, individuos, delimiter):
    # BASE - caso especial
    # inclui todas as variáveis constantes nas visitas
    start = time()
    print(f' - Escrevendo e compactando arquivo {BASE}',
          end='... ', flush=True)

    constantes = ('uf', 'capital', 'rm_ride', 'v1022', 'v1023',
                  'estrato', 'posest', 'v1014', 'pid', 'upa', 'v1008',
                  'pidgrp', 'pidgrpent', 'pidind', 'pidcla', 'pidindent',
                  'piddnd', 'piddnm', 'piddna')
    variaveis = [var for var in pesheader
                 if var[:2] in ('v1', 'v2') or var[:3] == 'vd2'
                 and var not in constantes]
    variaveis.sort()
    variaveis = ['ano', 'trimestre', 'pidpeso'] + variaveis
    header = constantes + tuple(f'{var}_{vis}' for vis in range(1, 6)
                                for var in variaveis)
    tgt = StringIO()
    tgt.write(delimiter.join(header) + '\n')
    for indkey in individuos:
        reg = [individuos[indkey][0][pesheader.index(var)]
               for var in constantes]
        track = 1
        for pessoa in individuos[indkey]:
            visita = int(pessoa[pesheader.index('v1016')])
            while visita > track:
                reg.extend([''] * len(variaveis))
                track += 1
            reg.extend([pessoa[pesheader.index(var)]
                        for var in variaveis])
            track += 1
        while track < 6:
            reg.extend([''] * len(variaveis))
            track += 1
        tgt.write(delimiter.join(reg) + '\n')
    print('OK!', __time__(start))
    tgt.seek(0)
    tgtname = f'{REGIND}/microdados.pnadc.paineis.{panel}.{BASE}.csv'
    with ZipFile(panelfile, 'a', ZIP_DEFLATED) as archive:
        archive.writestr(tgtname, tgt.read())
    del tgt


def __write_sections__(panel, panelfile, pesheader, individuos, delimiter):
    constantes = ('pid', 'upa', 'v1008', 'pidgrp', 'pidind')
    sections = ((EDUC, ('v3',)),
                (TRAB, ('v4',)),
                (OREN, ('v5', 'vi5')),
                (DERI, ('vd', 'vdi')),
                (MORA, ('s01',)),
                (TICS, ('s07',)),
                (TURI, ('s08',)),
                (TRIN, ('s06', 'sd')))
    for section in sections:
        variaveis = [var for var in pesheader
                     if any(prefix in var[:len(prefix)]
                            for prefix in section[1])]
        if not variaveis:
            continue
        variaveis.sort()
        start = time()
        print(f' - Escrevendo e compactando arquivo da seção {section[0]}',
              end='... ', flush=True)
        header = constantes + tuple(f'{var}_{vis}' for vis in range(1, 6)
                                    for var in variaveis)
        datafile = StringIO()
        datafile.write(delimiter.join(header) + '\n')
        for indkey in individuos:
            reg = [individuos[indkey][0][pesheader.index(var)]
                   for var in constantes]
            track = 1
            for pessoa in individuos[indkey]:
                visita = int(pessoa[pesheader.index('v1016')])
                while visita > track:
                    reg.extend([''] * len(variaveis))
                    track += 1
                reg.extend([pessoa[pesheader.index(var)]
                            for var in variaveis])
                track += 1
            while track < 6:
                reg.extend([''] * len(variaveis))
                track += 1
            datafile.write(delimiter.join(reg) + '\n')
        datafile = __drop_columns__(datafile)
        datafile.seek(0)
        tgtname = f'{REGIND}/microdados.pnadc.paineis.{panel}.{section[0]}.csv'
        with ZipFile(panelfile, 'a', ZIP_DEFLATED) as archive:
            archive.writestr(tgtname, datafile.read())
        del datafile
        print('OK!', __time__(start))
